const { args: contextArgs, normalizeGenerator, normalizeSemantic } = require('./helpers')
const Lines = require('../lines')

class Semantic {
  // constructor ({match, apply, uuid, index, km, notes}) {
  constructor (semantic) {
    semantic = normalizeSemantic(semantic)
    const { match, apply, uuid, index, km, notes } = semantic
    this.matcher = match
    this._apply = apply
    this.uuid = uuid
    this.index = index
    this.km = km
    this.notes = notes
  }

  toLabel () {
    const notes = (this.notes) ? ` notes: "${this.notes}"` : ''
    return `KM '${this.km}' ordinal: ${this.index} ${notes}`
  }

  toString () {
    return `Semantic(${this.matcher}, ${this._apply})`
  }

  getAPI (config) {
    if (config && config.getAPI) {
      return config.getAPI(this.uuid)
    }
  }

  matches (baseArgs, context) {
    const hierarchy = baseArgs.hierarchy
    const config = baseArgs.config
    const objects = baseArgs.getObjects(this.uuid)
    // return this.matcher(Object.assign({}, argsBase, {args: contextArgs(context, hierarchy), objects: objects, global: objects, context: context, hierarchy: hierarchy, api: this.getAPI(config)})
    const moreArgs = { args: contextArgs(context, hierarchy), objects: objects, global: objects, context: context, hierarchy: hierarchy, api: this.getAPI(config) }
    const args = Object.assign({}, baseArgs, moreArgs)
    return this.matcher(args)
  }

  apply (baseArgs, context, s, log) {
    const { hierarchy, config, response } = baseArgs
    const objects = baseArgs.getObjects(this.uuid)
    if (!log) {
      console.trace()
      throw 'log is a required argument'
    }
    const contextPrime = Object.assign({}, context)
    let n = (id) => id
    if (config && 'nsToString' in config) {
      n = (id) => config.nsToString(id)
    }
    const km = (name) => config.getConfig(name)
    const moreArgs = { km, args: contextArgs(context, hierarchy), objects, log, global: objects, n, hierarchy, context: contextPrime, uuid: this.uuid, config, response, s, api: this.getAPI(config) }
    const args = Object.assign({}, baseArgs, moreArgs)
    this._apply(args)
    return contextPrime
  }
}

class Semantics {
  constructor (semantics, logs = []) {
    this.semantics = semantics.map((semantic) => {
      if (semantic instanceof Semantic) {
        return semantic
      } else {
        return new Semantic(semantic)
      }
    })
    this.logs = logs
  };

  applyToContext (args, context) {
    // let context_prime = {}
    if (!(context instanceof Array || context instanceof Object)) {
      return context
    }

    const config = args.config
    let contextPrime = Object.assign({}, context)
    const s = (context) => this.apply(args, context)
    let applied = false
    for (const isemantic in this.semantics) {
      const semantic = this.semantics[isemantic]
      if (semantic.matches(args, context)) {
        const log = (message) => { this.logs.push(message) }
        contextPrime = semantic.apply(args, context, s, log)
        // this.logs.push(`Semantics: applied ${semantic.toString()}\n to\n ${JSON.stringify(context)}\n the result was ${JSON.stringify(contextPrime)}\n`)
        if (((config || {}).config || {}).debug) {
          const widths = [10, 10, 90]
          const lines = new Lines(widths)
          lines.setElement(0, 0, 'Semantic')
          if (semantic.index > -1 && semantic.km) {
            // lines.setElement(0, 2, `KM '${semantic.km}' ordinal: ${semantic.index} ${ semantic.notes ? semantic.nodes : '' }`)
            lines.setElement(0, 2, semantic.toLabel())
          }
          lines.newRow()
          lines.setElement(0, 1, 'APPLIED')
          lines.setElement(0, 2, semantic.toString())
          lines.newRow()
          lines.setElement(0, 1, 'TO')
          lines.setElement(0, 2, JSON.stringify(context, null, 2))
          lines.newRow()
          lines.setElement(0, 1, 'RESULT')
          lines.setElement(0, 2, JSON.stringify(contextPrime, null, 2))
          this.logs.push(lines.toString())
        }
        applied = true
        break
      }
    }
    if (!applied && ((config || {}).config || {}).debug) {
      const widths = [10, 10, 90]
      const lines = new Lines(widths)
      lines.setElement(0, 0, 'Semantic')
      lines.setElement(0, 2, 'No semantic applied')
      lines.newRow()
      lines.setElement(0, 1, 'TO')
      lines.setElement(0, 2, JSON.stringify(context, null, 2))
      this.logs.push(lines.toString())
    }
    return contextPrime
  }

  applyToContexts (args, contexts) {
    const contextsPrime = []
    contexts.forEach((context) => {
      contextsPrime.push(this.applyToContext(args, context))
    })
    return contextsPrime
  }

  apply (args, context) {
    if (Array.isArray(context)) {
      return this.applyToContexts(args, context)
    } else if (context instanceof Object) {
      return this.applyToContext(args, context)
    } else {
      return context
    }
  }
}

module.exports = { Semantic, Semantics, normalizeGenerator }
