const _ = require('lodash')
const deepEqual = require('deep-equal')

const groupBy = (property, list) => {
  const groups = {}
  for (const element of list) {
    if (!groups[element[property]]) {
      groups[element[property]] = []
    }
    groups[element[property]].push(element)
  }
  return groups
}

const toList = (value) => {
  if (value.marker === 'list') {
    return value
  } else {
    return {
      marker: 'list',
      types: Array.from(new Set((value.types || []).concat(['list', value.marker]))),
      value: [value]
    }
  }
}

const concatLists = (l1, l2) => {
  l1 = toList(l1)
  l2 = toList(l2)
  return {
    marker: 'list',
    types: Array.from(new Set(l1.types.concat(l2.types))),
    value: Array.from(new Set(l1.value.concat(l2.value)))
  }
}

const unflatten = (contexts) => {
  const groups = groupBy('marker', contexts)
  const unflats = []
  for (const key in groups) {
    const grouping = groups[key]
    const first = _.cloneDeep(grouping[0])
    for (const next of grouping.slice(1)) {
      for (const prop in next) {
        if (!deepEqual(next[prop], first[prop])) {
          first[prop] = concatLists(first[prop], next[prop])
        }
      }
    }
    unflats.push(first)
  }
  return unflats
}

module.exports = { groupBy, unflatten }
